# -*- coding: utf-8 -*-

import os
import signal
import sys

import boto3
import click
import requests
import six
import terminaltables

from dagster import seven

from .aws_util import (
    create_ec2_instance,
    create_key_pair,
    create_rds_instance,
    create_security_group,
    get_validated_ami_id,
    select_region,
    select_vpc,
)
from .config import HOST_CONFIG_FILE, EC2Config, RDSConfig
from .service_util import restart_dagit_service, stop_dagit_service
from .term import Spinner, Term, remove_ssh_key, rsync_to_remote, run_remote_cmd

# Client code will be deposited here on the remote EC2 instance
SERVER_CLIENT_CODE_HOME = '/opt/dagster/app/'

# DAGSTER_HOME on the remote instance
SERVER_DAGSTER_HOME = '/opt/dagster/dagster_home'

# Help text shown upon sync completion
COMPLETED_HELP_TEXT = '''🚀 To sync your Dagster project, in your project directory, run:

    dagster-aws up

You can also open a shell on your dagster-aws instance with:

    dagster-aws shell

For full details, see dagster-aws --help'''

DAGSTER_HOME_ERROR = '''DAGSTER_HOME is not set! Before continuing, set with e.g.:

export DAGSTER_HOME=~/.dagster

You may want to add this line to your .bashrc or .zshrc file.
'''

# We use the default PG port for the RDS instance
DEFAULT_RDS_PORT = 5432


def get_dagster_home():
    '''Ensures that the user has set a valid DAGSTER_HOME in environment and that it exists
    '''
    dagster_home_env = os.getenv('DAGSTER_HOME')
    if dagster_home_env is None:
        Term.fatal(DAGSTER_HOME_ERROR)

    dagster_home = os.path.expanduser(dagster_home_env)

    Term.info('Found DAGSTER_HOME in environment at: %s' % dagster_home)

    if not os.path.isdir(os.path.expanduser(dagster_home)):
        Term.fatal('The specified DAGSTER_HOME folder does not exist! Create before continuing.')
    return dagster_home


def exit_gracefully(_signum, _frame):
    '''Prevent stack trace spew on Ctrl-C
    '''
    click.echo(click.style('\n\nCommand killed by keyboard interrupt, quitting\n\n', fg='yellow'))
    sys.exit(1)


def sync_dagster_yaml(ec2_config, rds_config):
    '''Configure Dagster instance to use PG storage by putting a dagster.yaml file in the remote
    DAGSTER_HOME directory
    '''
    with open(os.path.join(os.path.dirname(__file__), 'conf', 'dagster.template.yaml'), 'rb') as f:
        dagster_yaml = six.ensure_str(f.read()).format(
            username=rds_config.username,
            password=rds_config.password,
            hostname=rds_config.instance_uri,
            db_name=rds_config.db_name,
            port=DEFAULT_RDS_PORT,
        )

    tmp_file = os.path.join(seven.get_system_temp_directory(), 'dagster.yaml')

    with open(tmp_file, 'wb') as f:
        f.write(six.ensure_binary(dagster_yaml))

    rsync_to_remote(ec2_config.key_file_path, tmp_file, ec2_config.remote_host, SERVER_DAGSTER_HOME)


def ensure_requirements(base_path):
    '''Check that the user has provided a requirements.txt, and if not, create one.
    '''
    requirements_file = os.path.join(base_path, 'requirements.txt')
    if os.path.exists(requirements_file):
        Term.info('Found existing requirements.txt')
        with open(requirements_file, 'rb') as f:
            reqs = six.ensure_str(f.read()).strip().split('\n')
        if not any([req.startswith('dagster') for req in reqs]):
            Term.warning('Could not find dagster in your requirements.txt')
            click.confirm('Continue?', default=False, abort=True)
        if not any([req.startswith('dagit') for req in reqs]):
            Term.warning('Could not find dagit in your requirements.txt')
            click.confirm('Continue?', default=False, abort=True)
    else:
        Term.waiting('No requirements.txt found, creating...')
        with open(requirements_file, 'wb') as f:
            f.write(six.ensure_binary('\n'.join(['dagster', 'dagit'])))


@click.group()
def main():
    signal.signal(signal.SIGINT, exit_gracefully)


@main.command()
@click.option('--use-master', is_flag=True, default=False)
def init(use_master):
    '''🚀 Initialize an EC2 VM to host Dagit'''
    click.echo('\n🌈 Welcome to Dagit + AWS quickstart cloud init!\n')

    # this ensures DAGSTER_HOME exists before we continue
    dagster_home = get_dagster_home()

    prev_config = None
    if EC2Config.exists(dagster_home):
        cfg_filepath = os.path.join(dagster_home, HOST_CONFIG_FILE)
        Term.warning('dagster-aws has already been initialized!\nState file:\n\t%s' % cfg_filepath)
        click.confirm('Continue?', default=False, abort=True)
        prev_config = EC2Config.load(dagster_home)

    region = select_region(prev_config)

    client = boto3.client('ec2', region_name=region)
    ec2 = boto3.resource('ec2', region_name=region)

    vpc = select_vpc(client, ec2)

    security_group_id = create_security_group(prev_config, client, ec2, vpc)

    ami_id = get_validated_ami_id(client)

    key_pair_name, key_file_path = create_key_pair(prev_config, client, dagster_home)

    inst = create_ec2_instance(client, ec2, security_group_id, ami_id, key_pair_name, use_master)

    # Save host configuration for future commands
    ec2_config = EC2Config(
        remote_host=inst.public_dns_name,
        instance_id=inst.id,
        region=region,
        security_group_id=security_group_id,
        key_pair_name=key_pair_name,
        key_file_path=key_file_path,
        ami_id=ami_id,
    )
    ec2_config.save(dagster_home)

    rds_config = create_rds_instance(dagster_home, region)

    click.echo(ec2_config.as_table() + '\n')

    if rds_config:
        rds_config.save(dagster_home)
        click.echo(rds_config.as_table() + '\n')
        sync_dagster_yaml(ec2_config, rds_config)

    click.echo(click.style(COMPLETED_HELP_TEXT, fg='green'))


@main.command()
def shell():
    '''Open an SSH shell on the remote server'''
    dagster_home = get_dagster_home()
    cfg = EC2Config.load(dagster_home)

    # Lands us directly in /opt/dagster (app dir may not exist yet)
    run_remote_cmd(cfg.key_file_path, cfg.remote_host, r'cd /opt/dagster; \$SHELL -l')


@main.command()
@click.option(
    '-p',
    '--post-up-script',
    type=click.Path(exists=True, dir_okay=False, resolve_path=True),
    help='Specify a path to a script with post-init actions',
)
def reconcile_scheduler_state(post_up_script):
    '''🌱 Sync your Dagster project to the remote server.

    This command will rsync the current folder to the remote host as /opt/dagster/app, and if there
    is a requirements.txt file provided in the current folder, it will install those requirements.
    '''
    dagster_home = get_dagster_home()
    cfg = EC2Config.load(dagster_home)

    if not cfg or not cfg.remote_host:
        Term.fatal('No EC2 config found; run dagster-aws init first!')

    if RDSConfig.exists(dagster_home):
        rds_config = RDSConfig.load(dagster_home)
        sync_dagster_yaml(cfg, rds_config)

    cwd = os.getcwd()

    if not os.path.exists(os.path.join(cwd, 'repository.yaml')):
        Term.fatal('No repository.yaml found in %s, create before continuing.' % cfg.local_path)

    if cfg.local_path is None:
        Term.info('Local path not configured; setting to %s' % cwd)
        cfg = cfg._replace(local_path=cwd)
        cfg.save(dagster_home)
    elif cfg.local_path != cwd:
        new_repo = click.confirm(
            'Local path already initialized! Sync a different repository?', default=False
        )
        if new_repo:
            stop_dagit_service(cfg)
            Term.info('Removing previous deployment from remote host...')
            run_remote_cmd(
                cfg.key_file_path, cfg.remote_host, 'rm -rf "%s*"' % SERVER_CLIENT_CODE_HOME
            )
            cfg = cfg._replace(local_path=cwd)
            cfg.save(dagster_home)

    ensure_requirements(cfg.local_path)

    rsync_to_remote(
        cfg.key_file_path,
        cfg.local_path + '/*',  # sync all files/directories in local_path
        cfg.remote_host,
        SERVER_CLIENT_CODE_HOME,
    )

    Term.waiting('Found a requirements.txt, ensuring dependencies are installed on remote host...')
    retval = run_remote_cmd(
        cfg.key_file_path,
        cfg.remote_host,
        'export PYTHONPATH=$PYTHONPATH:/opt/dagster/app && '
        'source /opt/dagster/venv/bin/activate && '
        'cd %s && pip install -r requirements.txt' % SERVER_CLIENT_CODE_HOME,
    )
    if retval == 0:
        Term.success('Install requirements.txt completed')
    else:
        Term.fatal('Error: could not install requirements.txt')

    setup_py_file = os.path.join(cfg.local_path, 'setup.py')
    if os.path.exists(setup_py_file):
        Term.waiting('Found a setup.py, installing on remote host...')
        with Spinner():
            retval = run_remote_cmd(
                cfg.key_file_path,
                cfg.remote_host,
                'export PYTHONPATH=$PYTHONPATH:/opt/dagster/app && '
                'source /opt/dagster/venv/bin/activate && '
                'cd %s && pip install .' % SERVER_CLIENT_CODE_HOME,
            )

    if post_up_script is not None:
        post_up_script = os.path.expanduser(post_up_script)

        Term.waiting('Copying user post-up script...')
        rsync_to_remote(cfg.key_file_path, post_up_script, cfg.remote_host, '/opt/dagster/')

        Term.waiting('Running user post-up script...')
        retval = run_remote_cmd(
            cfg.key_file_path,
            cfg.remote_host,
            'export PYTHONPATH=$PYTHONPATH:/opt/dagster/app && '
            'source /opt/dagster/venv/bin/activate && '
            'cd /opt/dagster/ && '
            'bash /opt/dagster/%s' % os.path.basename(post_up_script),
        )

    Term.waiting('Testing that pipeline loads correctly on remote host...')
    retval = run_remote_cmd(
        cfg.key_file_path,
        cfg.remote_host,
        'export PYTHONPATH=$PYTHONPATH:/opt/dagster/app && '
        'source /opt/dagster/venv/bin/activate && '
        'cd /opt/dagster/app/ && '
        '/opt/dagster/venv/bin/dagster pipeline list',
    )

    if retval == 0:
        Term.success('Pipeline test succeeded')
    else:
        Term.fatal('Errors in pipeline; fix before proceeding')

    restart_dagit_service(cfg)

    Term.success(
        'Synchronization succeeded. To open Dagit, visit the URL:\n\thttp://%s:3000'
        % cfg.remote_host
    )


@main.command()
def update_dagster():
    '''Update the remote copy of Dagster'''
    dagster_home = get_dagster_home()
    cfg = EC2Config.load(dagster_home)

    Term.waiting(
        'Running a git pull and make rebuild_dagit on the remote dagster, this may take a while...'
    )
    retval = run_remote_cmd(
        cfg.key_file_path, cfg.remote_host, 'cd /opt/dagster/dagster && git pull'
    )
    if retval == 0:
        Term.info('Dagster git refresh completed! Rebuilding dagit...')
    else:
        Term.fatal('git pull failed')

    retval = run_remote_cmd(
        cfg.key_file_path,
        cfg.remote_host,
        'source /opt/dagster/venv/bin/activate && '
        'cd /opt/dagster/dagster/ && '
        'make rebuild_dagit &&'
        'sudo service dagit restart',
    )
    if retval == 0:
        Term.success('Updating complete!')
    else:
        Term.fatal('Rebuilding dagit failed')


@main.command()
def delete():
    '''💥 Terminate your EC2 instance (and associated resources)'''
    dagster_home = get_dagster_home()

    already_run = EC2Config.exists(dagster_home)

    if not already_run:
        Term.fatal('No existing configuration detected, exiting')

    ec2_config = EC2Config.load(dagster_home)

    client = boto3.client('ec2', region_name=ec2_config.region)
    ec2 = boto3.resource('ec2', region_name=ec2_config.region)

    instances = ec2.instances.filter(  # pylint: disable=no-member
        InstanceIds=[ec2_config.instance_id]
    )

    Term.warning('This will terminate the following: ')
    for instance in instances:
        name = None
        for tag in instance.tags:
            if tag.get('Key') == 'Name':
                name = tag.get('Value')
        click.echo('\t%s %s %s' % (instance.id, name, instance.instance_type))

    click.confirm('\nThis step cannot be undone. Continue?', default=False, abort=True)

    Term.waiting('Terminating...')
    with Spinner():
        instances.terminate()
        waiter = client.get_waiter('instance_terminated')
        waiter.wait(InstanceIds=[instance.id for instance in instances])

    Term.rewind()
    Term.success('Done terminating instance')

    # Wipe all instance-related configs
    ec2_config = ec2_config._replace(
        remote_host=None, instance_id=None, ami_id=None, local_path=None
    )

    # Prompt user to remove key pair
    should_delete_key_pair = click.confirm(
        'Do you also want to remove the key pair %s?' % ec2_config.key_pair_name
    )
    if should_delete_key_pair:
        client.delete_key_pair(KeyName=ec2_config.key_pair_name)

        remove_ssh_key(ec2_config.key_file_path)

        ec2_config = ec2_config._replace(key_pair_name=None, key_file_path=None)

    # Prompt user to delete security group also
    should_remove_security_group = click.confirm(
        'Do you also want to remove the security group %s?' % ec2_config.security_group_id
    )
    if should_remove_security_group:
        client.delete_security_group(GroupId=ec2_config.security_group_id)
        ec2_config = ec2_config._replace(security_group_id=None)

    if should_delete_key_pair and should_remove_security_group:
        # Delete entirely
        ec2_config.delete(dagster_home)
    else:
        # Write out updated config
        ec2_config.save(dagster_home)

    # Prompt user to delete security group also
    if RDSConfig.exists(dagster_home):
        rds_config = RDSConfig.load(dagster_home)
        Term.warning('WARNING: A "yes" below will remove all RDS PostgreSQL data!!!')
        should_remove_rds = click.confirm(
            'Do you also want to remove the RDS instance %s?' % rds_config.instance_name
        )
        if should_remove_rds:
            rds = boto3.client('rds', region_name=ec2_config.region)
            rds.delete_db_instance(
                DBInstanceIdentifier=rds_config.instance_name,
                SkipFinalSnapshot=True,
                DeleteAutomatedBackups=True,
            )
            rds_config.delete(dagster_home)

    Term.success('Done!')


@main.command()
def info():
    '''Print out tabulated EC2/RDS configuration
    '''
    dagster_home = get_dagster_home()
    click.echo('\n')

    ec2_config = None
    if EC2Config.exists(dagster_home):
        ec2_config = EC2Config.load(dagster_home)
        click.echo(ec2_config.as_table() + '\n')
    else:
        Term.fatal('No EC2 instance configuration found; please run dagster-aws init first!')

    if RDSConfig.exists(dagster_home):
        rds_config = RDSConfig.load(dagster_home)
        click.echo(rds_config.as_table() + '\n')

    instance_uri = 'http://%s:3000' % ec2_config.remote_host

    click.echo('Instance status: ', nl=False)
    status = True
    try:
        r = requests.get(instance_uri, timeout=2.0)
        if r.status_code != 200:
            status = False
    except requests.RequestException:
        status = False

    if status:
        click.echo(click.style('[AVAILABLE]', fg='green'))
        click.echo(
            'To connect to your host, open the following URL in a browser:\n\n'
            + terminaltables.SingleTable(
                [['\n' + click.style('  %s  ' % instance_uri, fg='blue') + '\n']],
                title='Dagit Instance URI',
            ).table
        )

    else:
        click.echo(click.style('[DOWN]', fg='red'))
        Term.fatal('Could not connect to remote host %s, aborting' % instance_uri)



import copy
import os
import zipfile

from dagster import check


def subset_environment_dict(environment_dict, solid_name):
    '''Drops solid config for solids other than solid_name; this subsetting is required when
    executing a single solid on EMR to pass config validation.
    '''
    check.dict_param(environment_dict, 'environment_dict')
    check.str_param(solid_name, 'solid_name')

    subset = copy.deepcopy(environment_dict)
    if 'solids' in subset:
        solid_config_keys = list(subset['solids'].keys())
        for key in solid_config_keys:
            if key != solid_name:
                del subset['solids'][key]
    return subset


def build_pyspark_zip(zip_file, path):
    '''Archives the current path into a file named `zip_file`
    '''
    check.str_param(zip_file, 'zip_file')
    check.str_param(path, 'path')

    with zipfile.ZipFile(zip_file, 'w', zipfile.ZIP_DEFLATED) as zf:
        for root, _, files in os.walk(path):
            for fname in files:
                abs_fname = os.path.join(root, fname)

                # Skip various artifacts
                if 'pytest' in abs_fname or '__pycache__' in abs_fname or 'pyc' in abs_fname:
                    continue

                zf.write(abs_fname, os.path.relpath(os.path.join(root, fname), path))



# -*- coding: utf-8 -*-
import datetime
import getpass
import os
import time

import boto3
import click
import six
import terminaltables
from botocore.exceptions import ClientError

from .term import Spinner, Term

# Default to northern CA
DEFAULT_REGION = 'us-west-1'

# User can select an EC2 instance type if this is too small
DEFAULT_INSTANCE_TYPE = 't3.medium'

# Ubuntu Server 18.04 LTS (HVM), SSD Volume Type
DEFAULT_AMI = 'ami-08fd8ae3806f09a08'

# User can select an instance type for a PG RDS instance
DEFAULT_RDS_INSTANCE_TYPE = 'db.t3.small'

VPC_CREATION_WARNING = '''Note that this script will not create a VPC on your behalf. If you don't
know what this is, the default is almost certainly what you want. However,
please be aware that the instance we create in your VPC is likely to be
insecure (publicly accessible). For production, you will probably want to set
up a secured instance and VPC!
'''


def get_all_regions():
    '''Retrieves all regions/endpoints that work with EC2
    '''
    response = boto3.client('ec2').describe_regions()
    return sorted([r['RegionName'] for r in response['Regions']])


def select_region(prev_config):
    '''User can select the region in which to instantiate EC2 machine.
    '''
    if prev_config and prev_config.region:
        Term.success('Found existing region, continuing with %s' % prev_config.region)
        return prev_config.region

    all_regions = get_all_regions()

    table = terminaltables.SingleTable(
        [[r] for r in all_regions], title=click.style('Regions', fg='blue')
    )
    table.inner_heading_row_border = False
    click.echo('\n' + table.table)

    return click.prompt(
        'Select an AWS region ' + click.style('[default: %s]' % DEFAULT_REGION, fg='green'),
        type=click.Choice(all_regions),
        default=DEFAULT_REGION,
        show_default=False,
        show_choices=False,
    )


def select_vpc(client, ec2):
    '''User can select the VPC they would like to use for this machine.
    '''
    filters = [{'Name': 'isDefault', 'Values': ['true']}]
    default_vpcs = list(ec2.vpcs.filter(Filters=filters))

    # Will be either zero or one VPC; per boto3 docs,
    # "You cannot have more than one default VPC per Region."
    # https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ec2.html#EC2.Client.create_default_vpc
    if not default_vpcs:
        Term.fatal('AWS account does not have a default VPC; needs manual setup')

    default_vpc = default_vpcs[0].id

    vpcs = []
    for vpc in ec2.vpcs.iterator():
        response = client.describe_vpcs(VpcIds=[vpc.id])
        name = [tag['Value'] for tag in response['Vpcs'][0].get('Tags', []) if tag['Key'] == 'Name']
        if vpc.id == default_vpc:
            name = ['default']
        vpcs.append([vpc.id, name[0] if name else '<none>'])

    table = terminaltables.SingleTable(vpcs, title=click.style('Existing VPCs', fg='blue'))
    table.inner_heading_row_border = False
    click.echo('\n' + table.table)

    vpc = None
    vpc_ids = [vpc[0] for vpc in vpcs]

    Term.info(VPC_CREATION_WARNING)

    vpc = None
    while not vpc:
        vpc = click.prompt(
            '\nSelect an existing VPC ID to use. '
            + click.style('[default: %s]' % default_vpc, fg='green'),
            type=str,
            default=default_vpc,
            show_default=False,
        )
        if vpc not in vpc_ids:
            Term.error('Specified VPC %s does not exist' % vpc)
            vpc = None
    return vpc


def create_security_group(prev_config, client, ec2, vpc):
    '''Allows the user to select a security group, otherwise we create one for the dagit host.
    '''
    if prev_config and prev_config.security_group_id:
        Term.success(
            'Found existing security group, continuing with %s' % (prev_config.security_group_id)
        )
        return prev_config.security_group_id

    now = datetime.datetime.utcnow()
    default_group_name = 'dagster-sg-%s-%s' % (getpass.getuser(), now.strftime("%Y%m%dT%H%M%S"))

    existing_group_names = [
        g['GroupName']
        for g in client.describe_security_groups()['SecurityGroups']
        if g.get('VpcId') == vpc
    ]

    group_name = None
    while not group_name:
        group_name = click.prompt(
            '\nChoose a security group name '
            + click.style('[default is %s]' % default_group_name, fg='green'),
            type=str,
            default=default_group_name,
            show_default=False,
        )

        # Ensure key doesn't already exist
        if group_name in existing_group_names:
            Term.error('Specified security group already exists, won\'t create')
            group_name = None

        else:
            Term.waiting('Creating dagit security group...')
            group = ec2.create_security_group(
                GroupName=group_name, Description='dagit Security Group', VpcId=vpc
            )
            client.authorize_security_group_ingress(
                GroupId=group.id,
                IpPermissions=[
                    {
                        'IpProtocol': 'tcp',
                        'FromPort': 3000,
                        'ToPort': 3000,
                        'IpRanges': [{'CidrIp': '0.0.0.0/0', 'Description': 'HTTP'}],
                    },
                    {
                        'IpProtocol': 'tcp',
                        'FromPort': 22,
                        'ToPort': 22,
                        'IpRanges': [{'CidrIp': '0.0.0.0/0', 'Description': 'SSH'}],
                    },
                ],
            )
            group.create_tags(Tags=[{'Key': 'Name', 'Value': group_name}])
            Term.rewind()
            Term.success('Security group created!')
            return group.id


def create_key_pair(prev_config, client, dagster_home):
    '''Ensures we have a key pair in place for creating EC2 instance so that the user can SSH to the
    machine once it is created.
    '''
    if prev_config and prev_config.key_file_path and os.path.exists(prev_config.key_file_path):
        Term.success(
            'Found existing key pair %s, continuing with %s'
            % (prev_config.key_pair_name, prev_config.key_file_path)
        )
        return prev_config.key_pair_name, prev_config.key_file_path

    now = datetime.datetime.utcnow()
    default_key_pair_name = 'dagster-keypair-%s-%s' % (
        getpass.getuser(),
        now.strftime("%Y%m%dT%H%M%S"),
    )

    existing_key_pairs = [k['KeyName'] for k in client.describe_key_pairs()['KeyPairs']]

    key_pair_name = None
    while not key_pair_name:
        key_pair_name = click.prompt(
            '\nChoose a key pair name '
            + click.style('[default is %s]' % default_key_pair_name, fg='green'),
            type=str,
            default=default_key_pair_name,
            show_default=False,
        )

        # Ensure key doesn't already exist
        if key_pair_name in existing_key_pairs:
            Term.error('Specified key pair %s already exists, won\'t create' % key_pair_name)
            key_pair_name = None

        else:
            # key does not exist yet, safe to create
            Term.waiting('Creating new key pair...')
            keypair = client.create_key_pair(KeyName=key_pair_name)

            key_directory = os.path.join(dagster_home, 'keys')
            key_file_path = os.path.join(key_directory, '%s.pem' % key_pair_name)

            # Ensure key directory exists
            if not os.path.exists(key_directory):
                os.makedirs(key_directory)

            # Write key material and ensure appropriate permissions are set
            with open(key_file_path, 'wb') as f:
                f.write(six.ensure_binary(keypair['KeyMaterial']))
            os.chmod(key_file_path, 0o600)

            time.sleep(0.4)
            Term.rewind()
            Term.success(
                'Key pair %s created and saved to local file %s!' % (key_pair_name, key_file_path)
            )
            return six.ensure_str(key_pair_name), six.ensure_str(key_file_path)


def get_validated_ami_id(client):
    '''Permit the user to select an AMI to use. We default to using a Ubuntu 16.04 LTS AMI.
    '''

    ami_id = None
    while not ami_id:
        ami_id = click.prompt(
            '\nChoose an AMI to use (must be Debian-based) '
            + click.style('[default is %s (For us-west-1 region only)]' % DEFAULT_AMI, fg='green'),
            type=str,
            default=DEFAULT_AMI,
            show_default=False,
        )

        # Boto will throw a ClientError exception if this AMI doesn't exist.
        try:
            client.describe_images(ImageIds=[ami_id])
        except ClientError:
            Term.error('Specified AMI does not exist in the chosen region, fix to continue')
            ami_id = None

    return six.ensure_str(ami_id)


def create_ec2_instance(client, ec2, security_group_id, ami_id, key_pair_name, use_master):
    '''Actually create the EC2 instance given the provided configuration.
    '''
    click.echo(u'\n🌱 Provisioning an EC2 instance for dagit')

    instance_type = click.prompt(
        '\nChoose an EC2 instance type '
        + click.style('[default is %s]' % DEFAULT_INSTANCE_TYPE, fg='green'),
        type=str,
        default=DEFAULT_INSTANCE_TYPE,
        show_default=False,
    )

    ec2_default_name = 'dagit-ec2-%s' % getpass.getuser()
    ec2_instance_name = click.prompt(
        '\nChoose an EC2 instance name '
        + click.style('[default is %s]' % ec2_default_name, fg='green'),
        type=str,
        default=ec2_default_name,
        show_default=False,
    )

    init_script_name = 'init.sh' if use_master else 'init-stable.sh'

    # Here we actually create the EC2 instance
    Term.waiting('Creating dagit EC2 instance. This can take a couple of minutes...')
    with Spinner():
        with open(os.path.join(os.path.dirname(__file__), 'shell', init_script_name), 'rb') as f:
            init_script = six.ensure_str(f.read())

        instances = ec2.create_instances(
            ImageId=ami_id,
            InstanceType=instance_type,
            MaxCount=1,
            MinCount=1,
            SecurityGroupIds=[security_group_id],
            KeyName=key_pair_name,
            UserData=init_script,
        )
        inst = instances[0]
        waiter = client.get_waiter('instance_status_ok')
        waiter.wait(InstanceIds=[inst.id])

        # Add name tag to instance after creation
        ec2.create_tags(Resources=[inst.id], Tags=[{'Key': 'Name', 'Value': ec2_instance_name}])

        # Load details like public DNS name, etc.
        inst.reload()

    Term.rewind()
    Term.success('dagit EC2 instance %s launched!' % inst.id)

    return inst


def create_rds_instance(dagster_home, region):
    '''Creates an RDS PostgreSQL instance.

    Returns:
    RDSConfig object
    '''
    from .config import RDSConfig

    if RDSConfig.exists(dagster_home):
        prev_config = RDSConfig.load(dagster_home)
        Term.success('Found existing RDS database, continuing with %s' % (prev_config.instance_uri))
        return prev_config

    rds = boto3.client('rds', region_name=region)

    use_rds = click.confirm(
        '\nDo you want to use RDS PostgreSQL for run storage? (default: use filesystem) '
        + click.style('[y/N]', fg='green'),
        default=False,
        show_default=False,
    )
    if not use_rds:
        return None

    def generate_password(pass_len=16):
        client = boto3.client('secretsmanager')
        response = client.get_random_password(
            PasswordLength=pass_len, ExcludePunctuation=True, IncludeSpace=False
        )
        return response['RandomPassword']

    instance_type = click.prompt(
        '\nChoose an RDS instance type - see https://aws.amazon.com/rds/instance-types/ '
        + click.style('[default is %s]' % DEFAULT_RDS_INSTANCE_TYPE, fg='green'),
        type=str,
        default=DEFAULT_RDS_INSTANCE_TYPE,
        show_default=False,
    )

    default_name = 'dagster-rds-%s' % getpass.getuser()
    instance_name = click.prompt(
        '\nChoose an RDS instance name '
        + click.style('[default is %s]' % default_name, fg='green'),
        type=str,
        default=default_name,
        show_default=False,
    )

    rds_config = RDSConfig(
        instance_name=instance_name, instance_type=instance_type, password=generate_password()
    )

    Term.waiting(
        u'Creating Dagster RDS instance. This unfortunately takes quite a while, 5+ minutes, so '
        u'grab some 🍿...'
    )
    with Spinner():
        rds.create_db_instance(
            AllocatedStorage=rds_config.storage_size_gb,
            DBInstanceClass=rds_config.instance_type,
            DBInstanceIdentifier=rds_config.instance_name,
            DBName=rds_config.db_name,
            Engine=rds_config.db_engine,
            EngineVersion=rds_config.db_engine_version,
            MasterUsername=rds_config.username,
            MasterUserPassword=rds_config.password,
        )
        waiter_rds = rds.get_waiter('db_instance_available')
        waiter_rds.wait(DBInstanceIdentifier=instance_name)
        response = rds.describe_db_instances(DBInstanceIdentifier=instance_name)
        instance_uri = response['DBInstances'][0]['Endpoint']['Address']

    Term.rewind()
    Term.success('Dagster RDS instance %s launched!' % instance_uri)

    return rds_config._replace(instance_uri=instance_uri)



import os
import subprocess
import sys

import pytest
from dagster_aws.emr import EmrJobRunner
from dagster_aws.emr.emr_pyspark_step_launcher import emr_pyspark_step_launcher
from dagster_aws.s3 import s3_plus_default_storage_defs, s3_resource
from dagster_pyspark import DataFrame, pyspark_resource
from moto import mock_emr
from pyspark.sql import Row
from pyspark.sql.types import IntegerType, StringType, StructField, StructType

from dagster import (
    InputDefinition,
    ModeDefinition,
    OutputDefinition,
    execute_pipeline,
    pipeline,
    solid,
)
from dagster.core.definitions.handle import ExecutionTargetHandle
from dagster.core.definitions.no_step_launcher import no_step_launcher
from dagster.core.errors import DagsterSubprocessError
from dagster.seven import mock
from dagster.utils.merger import deep_merge_dicts
from dagster.utils.test import create_test_pipeline_execution_context

S3_BUCKET = 'dagster-scratch-80542c2'

BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG = {
    'local_pipeline_package_path': os.path.abspath(os.path.dirname(__file__)),
    'cluster_id': os.environ.get('EMR_CLUSTER_ID'),
    'staging_bucket': S3_BUCKET,
    'region_name': 'us-west-1',
    'deploy_local_pipeline_package': True,
}


@solid(
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher', 'pyspark'},
)
def make_df_solid(context):
    schema = StructType([StructField('name', StringType()), StructField('age', IntegerType())])
    rows = [Row(name='John', age=19), Row(name='Jennifer', age=29), Row(name='Henry', age=50)]
    return context.resources.pyspark.spark_session.createDataFrame(rows, schema)


@solid(
    name='blah',
    description='this is a test',
    config={'foo': str, 'bar': int},
    input_defs=[InputDefinition('people', DataFrame)],
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher'},
)
def filter_df_solid(_, people):
    return people.filter(people['age'] < 30)


MODE_DEFS = [
    ModeDefinition(
        'prod',
        resource_defs={
            'pyspark_step_launcher': emr_pyspark_step_launcher,
            'pyspark': pyspark_resource,
            's3': s3_resource,
        },
        system_storage_defs=s3_plus_default_storage_defs,
    ),
    ModeDefinition(
        'local',
        resource_defs={'pyspark_step_launcher': no_step_launcher, 'pyspark': pyspark_resource},
    ),
]


@pipeline(mode_defs=MODE_DEFS)
def pyspark_pipe():
    filter_df_solid(make_df_solid())


def define_pyspark_pipe():
    return pyspark_pipe


@solid(
    output_defs=[OutputDefinition(DataFrame)],
    required_resource_keys={'pyspark_step_launcher', 'pyspark'},
)
def do_nothing_solid(_):
    pass


@pipeline(mode_defs=MODE_DEFS)
def do_nothing_pipe():
    do_nothing_solid()


def define_do_nothing_pipe():
    return do_nothing_pipe


def test_local():
    result = execute_pipeline(
        pipeline=pyspark_pipe,
        mode='local',
        environment_dict={'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}}},
    )
    assert result.success


@mock_emr
@mock.patch('dagster_aws.emr.emr.EmrJobRunner.wait_for_emr_steps_to_complete')
@mock.patch('dagster_aws.emr.emr_pyspark_step_launcher.EmrPySparkStepLauncher.get_step_events')
def test_pyspark_emr(mock_wait, mock_get_step_events):
    run_job_flow_args = dict(
        Instances={
            'InstanceCount': 1,
            'KeepJobFlowAliveWhenNoSteps': True,
            'MasterInstanceType': 'c3.medium',
            'Placement': {'AvailabilityZone': 'us-west-1a'},
            'SlaveInstanceType': 'c3.xlarge',
        },
        JobFlowRole='EMR_EC2_DefaultRole',
        LogUri='s3://mybucket/log',
        Name='cluster',
        ServiceRole='EMR_DefaultRole',
        VisibleToAllUsers=True,
    )

    # Doing cluster setup outside of a solid here, because run_job_flow is not yet plumbed through
    # to the pyspark EMR resource.
    job_runner = EmrJobRunner(region='us-west-1')
    context = create_test_pipeline_execution_context()
    cluster_id = job_runner.run_job_flow(context.log, run_job_flow_args)

    pipeline_def = ExecutionTargetHandle.for_pipeline_fn(
        define_do_nothing_pipe
    ).build_pipeline_definition()
    result = execute_pipeline(
        pipeline=pipeline_def,
        mode='prod',
        environment_dict={
            'resources': {
                'pyspark_step_launcher': {
                    'config': deep_merge_dicts(
                        BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG, {'cluster_id': cluster_id}
                    ),
                }
            },
        },
    )
    assert result.success
    assert mock_wait.called_once
    assert mock_get_step_events.called_once


def sync_code():
    # Sync remote dagster packages with local dagster code
    sync_code_command = [
        'rsync',
        '-av',
        '-progress',
        "--exclude='scala_modules/'",
        "--exclude='js_modules/'",
        "--exclude='.git/'",
        "--exclude='docs/'",
        '-e',
        '"ssh -i {aws_emr_pem_file}"'.format(aws_emr_pem_file=os.environ['AWS_EMR_PEM_FILE']),
        os.environ['DAGSTER_DIR'],
        os.environ['AWS_EMR_NODE_ADDRESS'] + ':~/',
    ]
    if (
        subprocess.call(
            ' '.join(sync_code_command), stdout=sys.stdout, stderr=sys.stderr, shell=True
        )
        != 0
    ):
        raise DagsterSubprocessError('Failed to sync code to EMR')

    # Install dagster packages on remote node
    remote_install_dagster_packages_command = ['sudo', 'python3', '-m', 'pip', 'install'] + [
        token
        for package_subpath in ['dagster', 'libraries/dagster-pyspark']
        for token in ['-e', '/home/hadoop/dagster/python_modules/' + package_subpath]
    ]

    install_dagster_packages_command = [
        'ssh',
        '-i',
        os.environ['AWS_EMR_PEM_FILE'],
        os.environ['AWS_EMR_NODE_ADDRESS'],
        "'" + ' '.join(remote_install_dagster_packages_command) + "'",
    ]
    if (
        subprocess.call(
            ' '.join(install_dagster_packages_command),
            stdout=sys.stdout,
            stderr=sys.stderr,
            shell=True,
        )
        != 0
    ):
        raise DagsterSubprocessError('Failed to install dagster packages on EMR')


@pytest.mark.skipif(
    'AWS_EMR_TEST_DO_IT_LIVE' not in os.environ,
    reason='This test is slow and requires a live EMR cluster; run only upon explicit request',
)
def test_do_it_live_emr():
    sync_code()

    # Retrieving the pipeline this way stores pipeline definition in the ExecutionTargetHandle
    # cache, where it can be retrieved and sent to the remote cluster at launch time.
    pipeline_def = ExecutionTargetHandle.for_pipeline_fn(
        define_pyspark_pipe
    ).build_pipeline_definition()

    result = execute_pipeline(
        pipeline_def,
        mode='prod',
        environment_dict={
            'solids': {'blah': {'config': {'foo': 'a string', 'bar': 123}}},
            'resources': {
                'pyspark_step_launcher': {'config': BASE_EMR_PYSPARK_STEP_LAUNCHER_CONFIG},
            },
            'storage': {'s3': {'config': {'s3_bucket': S3_BUCKET, 's3_prefix': 'test_pyspark'}}},
        },
    )
    assert result.success



from dagster import (
    EventMetadataEntry,
    Field,
    FileHandle,
    InputDefinition,
    Materialization,
    Output,
    OutputDefinition,
    String,
    check,
    input_hydration_config,
    solid,
)
from dagster.core.types.dagster_type import PythonObjectDagsterType

from .file_manager import S3FileHandle


def dict_with_fields(name, fields):
    check.str_param(name, 'name')
    check.dict_param(fields, 'fields', key_type=str)
    field_names = set(fields.keys())

    @input_hydration_config(fields)
    def _input_schema(_context, value):
        check.dict_param(value, 'value')
        check.param_invariant(set(value.keys()) == field_names, 'value')
        return value

    class _DictWithSchema(PythonObjectDagsterType):
        def __init__(self):
            super(_DictWithSchema, self).__init__(
                python_type=dict, name=name, input_hydration_config=_input_schema
            )

    return _DictWithSchema()


S3Coordinate = dict_with_fields(
    'S3Coordinate',
    fields={
        'bucket': Field(String, description='S3 bucket name'),
        'key': Field(String, description='S3 key name'),
    },
)


def last_key(key):
    if '/' not in key:
        return key
    comps = key.split('/')
    return comps[-1]


@solid(
    config={
        'Bucket': Field(
            String, description='The name of the bucket to upload to.', is_required=True
        ),
        'Key': Field(String, description='The name of the key to upload to.', is_required=True),
    },
    input_defs=[InputDefinition('file_handle', FileHandle, description='The file to upload.')],
    output_defs=[OutputDefinition(name='s3_file_handle', dagster_type=S3FileHandle)],
    description='''Take a file handle and upload it to s3. Returns an S3FileHandle.''',
    required_resource_keys={'s3'},
)
def file_handle_to_s3(context, file_handle):
    bucket = context.solid_config['Bucket']
    key = context.solid_config['Key']

    with context.file_manager.read(file_handle, 'rb') as fileobj:
        context.resources.s3.upload_fileobj(fileobj, bucket, key)
        s3_file_handle = S3FileHandle(bucket, key)

        yield Materialization(
            label='file_to_s3',
            metadata_entries=[EventMetadataEntry.path(s3_file_handle.s3_path, label=last_key(key))],
        )

        yield Output(value=s3_file_handle, output_name='s3_file_handle')



import os
import pickle

import boto3
from dagster_aws.emr import EmrJobRunner, emr_step_main
from dagster_aws.emr.utils import build_pyspark_zip
from dagster_aws.s3.file_manager import S3FileHandle
from dagster_aws.utils.mrjob.log4j import parse_hadoop_log4j_records
from dagster_spark.configs_spark import spark_config as get_spark_config
from dagster_spark.utils import flatten_dict, format_for_cli

from dagster import Field, StringSource, check, resource, seven
from dagster.core.definitions.step_launcher import StepLauncher
from dagster.core.events import log_step_event
from dagster.core.execution.plan.external_step import (
    PICKLED_EVENTS_FILE_NAME,
    PICKLED_STEP_RUN_REF_FILE_NAME,
    step_context_to_step_run_ref,
)

# On EMR, Spark is installed here
EMR_SPARK_HOME = '/usr/lib/spark/'

CODE_ZIP_NAME = 'code.zip'


@resource(
    {
        'spark_config': get_spark_config(),
        'cluster_id': Field(
            StringSource, description='Name of the job flow (cluster) on which to execute'
        ),
        'region_name': Field(StringSource),
        'action_on_failure': Field(str, is_required=False, default_value='CANCEL_AND_WAIT'),
        'staging_bucket': Field(
            StringSource,
            is_required=True,
            description='S3 bucket to use for passing files between the plan process and EMR '
            'process.',
        ),
        'staging_prefix': Field(
            StringSource,
            is_required=False,
            default_value='emr_staging',
            description='S3 key prefix inside the staging_bucket to use for files passed the plan '
            'process and EMR process',
        ),
        'wait_for_logs': Field(
            bool,
            is_required=False,
            default_value=False,
            description='If set, the system will wait for EMR logs to appear on S3. Note that logs '
            'are copied every 5 minutes, so enabling this will add several minutes to the job '
            'runtime.',
        ),
        'local_pipeline_package_path': Field(
            StringSource,
            is_required=True,
            description='Absolute path to the package that contains the pipeline definition(s) '
            'whose steps will execute remotely on EMR. This is a path on the local fileystem of '
            'the process executing the pipeline. The expectation is that this package will also be '
            'available on the python path of the launched process running the Spark step on EMR, '
            'either deployed on step launch via the deploy_pipeline_package option, referenced on '
            's3 via the s3_pipeline_package_path option, or installed on the cluster via bootstrap '
            'actions.',
        ),
        'deploy_local_pipeline_package': Field(
            bool,
            default_value=False,
            is_required=False,
            description='If set, before every step run, the launcher will zip up all the code in '
            'local_pipeline_package_path, upload it to s3, and pass it to spark-submit\'s '
            '--py-files option. This gives the remote process access to up-to-date user code. '
            'If not set, the assumption is that some other mechanism is used for distributing code '
            'to the EMR cluster. If this option is set to True, s3_pipeline_package_path should '
            'not also be set.',
        ),
        's3_pipeline_package_path': Field(
            StringSource,
            is_required=False,
            description='If set, this path will be passed to the --py-files option of spark-submit. '
            'This should usually be a path to a zip file.  If this option is set, '
            'deploy_local_pipeline_package should not be set to True.',
        ),
    }
)
def emr_pyspark_step_launcher(context):
    return EmrPySparkStepLauncher(**context.resource_config)


class EmrPySparkStepLauncher(StepLauncher):
    def __init__(
        self,
        region_name,
        staging_bucket,
        staging_prefix,
        wait_for_logs,
        action_on_failure,
        cluster_id,
        spark_config,
        local_pipeline_package_path,
        deploy_local_pipeline_package,
        s3_pipeline_package_path=None,
    ):
        self.region_name = check.str_param(region_name, 'region_name')
        self.staging_bucket = check.str_param(staging_bucket, 'staging_bucket')
        self.staging_prefix = check.str_param(staging_prefix, 'staging_prefix')
        self.wait_for_logs = check.bool_param(wait_for_logs, 'wait_for_logs')
        self.action_on_failure = check.str_param(action_on_failure, 'action_on_failure')
        self.cluster_id = check.str_param(cluster_id, 'cluster_id')
        self.spark_config = spark_config

        check.invariant(
            not deploy_local_pipeline_package or not s3_pipeline_package_path,
            'If deploy_local_pipeline_package is set to True, s3_pipeline_package_path should not '
            'also be set.',
        )

        self.local_pipeline_package_path = check.str_param(
            local_pipeline_package_path, 'local_pipeline_package_path'
        )
        self.deploy_local_pipeline_package = check.bool_param(
            deploy_local_pipeline_package, 'deploy_local_pipeline_package'
        )
        self.s3_pipeline_package_path = check.opt_str_param(
            s3_pipeline_package_path, 's3_pipeline_package_path'
        )

        self.emr_job_runner = EmrJobRunner(region=self.region_name)

    def _post_artifacts(self, log, step_run_ref, run_id, step_key):
        '''
        Synchronize the step run ref and pyspark code to an S3 staging bucket for use on EMR.

        For the zip file, consider the following toy example:

            # Folder: my_pyspark_project/
            # a.py
            def foo():
                print(1)

            # b.py
            def bar():
                print(2)

            # main.py
            from a import foo
            from b import bar

            foo()
            bar()

        This will zip up `my_pyspark_project/` as `my_pyspark_project.zip`. Then, when running
        `spark-submit --py-files my_pyspark_project.zip emr_step_main.py` on EMR this will
        print 1, 2.
        '''

        with seven.TemporaryDirectory() as temp_dir:
            s3 = boto3.client('s3', region_name=self.region_name)

            # Upload step run ref
            def _upload_file_to_s3(local_path, s3_filename):
                key = self._artifact_s3_key(run_id, step_key, s3_filename)
                s3_uri = self._artifact_s3_uri(run_id, step_key, s3_filename)
                log.debug(
                    'Uploading file {local_path} to {s3_uri}'.format(
                        local_path=local_path, s3_uri=s3_uri
                    )
                )
                s3.upload_file(Filename=local_path, Bucket=self.staging_bucket, Key=key)

            # Upload main file.
            # The remote Dagster installation should also have the file, but locating it there
            # could be a pain.
            main_local_path = self._main_file_local_path()
            _upload_file_to_s3(main_local_path, self._main_file_name())

            if self.deploy_local_pipeline_package:
                # Zip and upload package containing pipeline
                zip_local_path = os.path.join(temp_dir, CODE_ZIP_NAME)
                build_pyspark_zip(zip_local_path, self.local_pipeline_package_path)
                _upload_file_to_s3(zip_local_path, CODE_ZIP_NAME)

            # Create step run ref pickle file
            step_run_ref_local_path = os.path.join(temp_dir, PICKLED_STEP_RUN_REF_FILE_NAME)
            with open(step_run_ref_local_path, 'wb') as step_pickle_file:
                pickle.dump(step_run_ref, step_pickle_file)

            _upload_file_to_s3(step_run_ref_local_path, PICKLED_STEP_RUN_REF_FILE_NAME)

    def launch_step(self, step_context, prior_attempts_count):
        step_run_ref = step_context_to_step_run_ref(
            step_context, prior_attempts_count, self.local_pipeline_package_path
        )

        run_id = step_context.pipeline_run.run_id
        log = step_context.log

        step_key = step_run_ref.step_key
        self._post_artifacts(log, step_run_ref, run_id, step_key)

        emr_step_def = self._get_emr_step_def(run_id, step_key, step_context.solid.name)
        emr_step_id = self.emr_job_runner.add_job_flow_steps(log, self.cluster_id, [emr_step_def])[
            0
        ]
        self.emr_job_runner.wait_for_emr_steps_to_complete(log, self.cluster_id, [emr_step_id])
        if self.wait_for_logs:
            self._log_logs_from_s3(log, emr_step_id)

        for event in self.get_step_events(step_context, run_id, step_key):
            log_step_event(step_context, event)
            yield event

    def get_step_events(self, step_context, run_id, step_key):
        events_file_handle = S3FileHandle(
            self.staging_bucket, self._artifact_s3_key(run_id, step_key, PICKLED_EVENTS_FILE_NAME),
        )
        events_data = step_context.file_manager.read_data(events_file_handle)
        return pickle.loads(events_data)

    def _log_logs_from_s3(self, log, emr_step_id):
        '''Retrieves the logs from the remote PySpark process that EMR posted to S3 and logs
        them to the given log.'''
        stdout_log, stderr_log = self.emr_job_runner.retrieve_logs_for_step_id(
            log, self.cluster_id, emr_step_id
        )
        # Since stderr is YARN / Hadoop Log4J output, parse and reformat those log lines for
        # Dagster's logging system.
        records = parse_hadoop_log4j_records(stderr_log)
        for record in records:
            log._log(  # pylint: disable=protected-access
                record.level, record.logger + ': ' + record.message, {}
            )
        log.info(stdout_log)

    def _get_emr_step_def(self, run_id, step_key, solid_name):
        '''From the local Dagster instance, construct EMR steps that will kick off execution on a
        remote EMR cluster.
        '''
        action_on_failure = self.action_on_failure

        # Execute Solid via spark-submit
        conf = dict(flatten_dict(self.spark_config))
        conf['spark.app.name'] = conf.get('spark.app.name', solid_name)

        check.invariant(
            conf.get('spark.master', 'yarn') == 'yarn',
            desc='spark.master is configured as %s; cannot set Spark master on EMR to anything '
            'other than "yarn"' % conf.get('spark.master'),
        )

        command = (
            [
                EMR_SPARK_HOME + 'bin/spark-submit',
                '--master',
                'yarn',
                '--deploy-mode',
                conf.get('spark.submit.deployMode', 'client'),
            ]
            + format_for_cli(list(flatten_dict(conf)))
            + [
                '--py-files',
                self._artifact_s3_uri(run_id, step_key, CODE_ZIP_NAME),
                self._artifact_s3_uri(run_id, step_key, self._main_file_name()),
                self.staging_bucket,
                self._artifact_s3_key(run_id, step_key, PICKLED_STEP_RUN_REF_FILE_NAME),
            ]
        )

        return EmrJobRunner.construct_step_dict_for_command(
            'Execute Solid %s' % solid_name, command, action_on_failure=action_on_failure
        )

    def _main_file_name(self):
        return os.path.basename(self._main_file_local_path())

    def _main_file_local_path(self):
        return emr_step_main.__file__

    def _artifact_s3_uri(self, run_id, step_key, filename):
        key = self._artifact_s3_key(run_id, step_key, filename)
        return 's3://{bucket}/{key}'.format(bucket=self.staging_bucket, key=key)

    def _artifact_s3_key(self, run_id, step_key, filename):
        return '/'.join([self.staging_prefix, run_id, step_key, os.path.basename(filename)])



import os
import uuid

import yaml
from dagster_aws.cli.config import HOST_CONFIG_FILE, EC2Config, RDSConfig

from dagster import seven


def test_write_configs():
    ec2_config = EC2Config(
        remote_host='foo',
        region='us-west-1',
        security_group_id='sg-12345',
        key_pair_name='foobar',
        key_file_path='/some/path',
        ami_id='ami-12345',
    )

    rds_config = RDSConfig(
        instance_name='foo', instance_uri='foo-bar.amazonaws.com', password='baz'
    )

    # Ensure unique dir for test
    tmp_dir = os.path.join(seven.get_system_temp_directory(), uuid.uuid4().hex)
    outfile = os.path.join(tmp_dir, HOST_CONFIG_FILE)

    ec2_config.save(tmp_dir)
    rds_config.save(tmp_dir)

    with open(outfile) as f:
        record = yaml.safe_load(f)

    ec2_config_dict = record['ec2']
    rds_config_dict = record['rds']

    assert ec2_config_dict['remote_host'] == 'foo'
    assert ec2_config_dict['region'] == 'us-west-1'
    assert ec2_config_dict['security_group_id'] == 'sg-12345'
    assert ec2_config_dict['key_pair_name'] == 'foobar'
    assert ec2_config_dict['key_file_path'] == '/some/path'
    assert ec2_config_dict['ami_id'] == 'ami-12345'
    assert EC2Config.load(tmp_dir) == EC2Config(**ec2_config_dict) == ec2_config

    assert rds_config_dict['instance_name'] == 'foo'
    assert rds_config_dict['instance_uri'] == 'foo-bar.amazonaws.com'
    assert rds_config_dict['storage_size_gb'] == 20
    assert rds_config_dict['db_engine'] == 'postgres'
    assert rds_config_dict['db_engine_version'] == '11.6'
    assert RDSConfig.load(tmp_dir) == RDSConfig(**rds_config_dict)

    # Delete both configs
    res = rds_config.delete(tmp_dir)
    assert res
    assert not RDSConfig.exists(tmp_dir)
    res = ec2_config.delete(tmp_dir)
    assert res
    assert not EC2Config.exists(tmp_dir)

    # Try to delete non-existent config
    res = rds_config.delete(tmp_dir)
    assert not res


def test_table():
    cfg = EC2Config(
        remote_host='foo',
        region='us-west-1',
        security_group_id='sg-12345',
        key_pair_name='foobar',
        key_file_path='/some/path',
        ami_id='ami-12345',
    )

    cfg_table = cfg.as_table()
    assert 'EC2 Configuration' in cfg_table

    rds_config = RDSConfig(
        instance_name='foo', instance_uri='foo-bar.amazonaws.com', password='top secret'
    )
    rds_config_table = rds_config.as_table()
    assert 'RDS Configuration' in rds_config_table
    assert 'top secret' not in rds_config_table
    assert '<redacted>' in rds_config_table



import csv
import os
from collections import OrderedDict

import pytest
from dagster_aws.s3 import S3IntermediateStore, s3_plus_default_storage_defs, s3_resource

from dagster import (
    Bool,
    InputDefinition,
    Int,
    List,
    ModeDefinition,
    OutputDefinition,
    PipelineRun,
    SerializationStrategy,
    String,
    check,
    execute_pipeline,
    lambda_solid,
    pipeline,
    usable_as_dagster_type,
)
from dagster.core.events import DagsterEventType
from dagster.core.execution.api import create_execution_plan, execute_plan, scoped_pipeline_context
from dagster.core.instance import DagsterInstance
from dagster.core.storage.type_storage import TypeStoragePlugin, TypeStoragePluginRegistry
from dagster.core.types.dagster_type import Bool as RuntimeBool
from dagster.core.types.dagster_type import String as RuntimeString
from dagster.core.types.dagster_type import create_any_type, resolve_dagster_type
from dagster.core.utils import make_new_run_id
from dagster.utils.test import yield_empty_pipeline_context


class UppercaseSerializationStrategy(SerializationStrategy):  # pylint: disable=no-init
    def serialize(self, value, write_file_obj):
        return write_file_obj.write(bytes(value.upper().encode('utf-8')))

    def deserialize(self, read_file_obj):
        return read_file_obj.read().decode('utf-8').lower()


LowercaseString = create_any_type(
    'LowercaseString', serialization_strategy=UppercaseSerializationStrategy('uppercase'),
)


def aws_credentials_present():
    return os.getenv('AWS_ACCESS_KEY_ID') and os.getenv('AWS_SECRET_ACCESS_KEY')


nettest = pytest.mark.nettest


def define_inty_pipeline(should_throw=True):
    @lambda_solid
    def return_one():
        return 1

    @lambda_solid(input_defs=[InputDefinition('num', Int)], output_def=OutputDefinition(Int))
    def add_one(num):
        return num + 1

    @lambda_solid
    def user_throw_exception():
        raise Exception('whoops')

    @pipeline(
        mode_defs=[
            ModeDefinition(
                system_storage_defs=s3_plus_default_storage_defs, resource_defs={'s3': s3_resource}
            )
        ]
    )
    def basic_external_plan_execution():
        add_one(return_one())
        if should_throw:
            user_throw_exception()

    return basic_external_plan_execution


def get_step_output(step_events, step_key, output_name='result'):
    for step_event in step_events:
        if (
            step_event.event_type == DagsterEventType.STEP_OUTPUT
            and step_event.step_key == step_key
            and step_event.step_output_data.output_name == output_name
        ):
            return step_event
    return None


@nettest
def test_using_s3_for_subplan(s3_bucket):
    pipeline_def = define_inty_pipeline()

    environment_dict = {'storage': {'s3': {'config': {'s3_bucket': s3_bucket}}}}

    run_id = make_new_run_id()

    execution_plan = create_execution_plan(pipeline_def, environment_dict=environment_dict)

    assert execution_plan.get_step_by_key('return_one.compute')

    step_keys = ['return_one.compute']
    instance = DagsterInstance.ephemeral()
    pipeline_run = PipelineRun(
        pipeline_name=pipeline_def.name, run_id=run_id, environment_dict=environment_dict
    )

    return_one_step_events = list(
        execute_plan(
            execution_plan.build_subset_plan(step_keys),
            environment_dict=environment_dict,
            pipeline_run=pipeline_run,
            instance=instance,
        )
    )

    assert get_step_output(return_one_step_events, 'return_one.compute')
    with scoped_pipeline_context(
        execution_plan.build_subset_plan(['return_one.compute']),
        environment_dict,
        pipeline_run,
        instance,
    ) as context:

        store = S3IntermediateStore(
            s3_bucket,
            run_id,
            s3_session=context.scoped_resources_builder.build(required_resource_keys={'s3'},).s3,
        )
        assert store.has_intermediate(context, 'return_one.compute')
        assert store.get_intermediate(context, 'return_one.compute', Int).obj == 1

    add_one_step_events = list(
        execute_plan(
            execution_plan.build_subset_plan(['add_one.compute']),
            environment_dict=environment_dict,
            pipeline_run=pipeline_run,
            instance=instance,
        )
    )

    assert get_step_output(add_one_step_events, 'add_one.compute')
    with scoped_pipeline_context(
        execution_plan.build_subset_plan(['add_one.compute']),
        environment_dict,
        pipeline_run,
        instance,
    ) as context:
        assert store.has_intermediate(context, 'add_one.compute')
        assert store.get_intermediate(context, 'add_one.compute', Int).obj == 2


class FancyStringS3TypeStoragePlugin(TypeStoragePlugin):  # pylint:disable=no-init
    @classmethod
    def compatible_with_storage_def(cls, _):
        # Not needed for these tests
        raise NotImplementedError()

    @classmethod
    def set_object(cls, intermediate_store, obj, context, dagster_type, paths):
        check.inst_param(intermediate_store, 'intermediate_store', S3IntermediateStore)
        paths.append(obj)
        return intermediate_store.set_object('', context, dagster_type, paths)

    @classmethod
    def get_object(cls, intermediate_store, _context, _dagster_type, paths):
        check.inst_param(intermediate_store, 'intermediate_store', S3IntermediateStore)
        res = intermediate_store.object_store.s3.list_objects(
            Bucket=intermediate_store.object_store.bucket,
            Prefix=intermediate_store.key_for_paths(paths),
        )
        return res['Contents'][0]['Key'].split('/')[-1]


@nettest
def test_s3_intermediate_store_with_type_storage_plugin(s3_bucket):
    run_id = make_new_run_id()

    intermediate_store = S3IntermediateStore(
        run_id=run_id,
        s3_bucket=s3_bucket,
        type_storage_plugin_registry=TypeStoragePluginRegistry(
            [(RuntimeString, FancyStringS3TypeStoragePlugin)]
        ),
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_value('hello', context, RuntimeString, ['obj_name'])

            assert intermediate_store.has_object(context, ['obj_name'])
            assert intermediate_store.get_value(context, RuntimeString, ['obj_name']) == 'hello'

        finally:
            intermediate_store.rm_object(context, ['obj_name'])


@nettest
def test_s3_intermediate_store_with_composite_type_storage_plugin(s3_bucket):
    run_id = make_new_run_id()

    intermediate_store = S3IntermediateStore(
        run_id=run_id,
        s3_bucket=s3_bucket,
        type_storage_plugin_registry=TypeStoragePluginRegistry(
            [(RuntimeString, FancyStringS3TypeStoragePlugin)]
        ),
    )

    with yield_empty_pipeline_context(run_id=run_id) as context:
        with pytest.raises(check.NotImplementedCheckError):
            intermediate_store.set_value(
                ['hello'], context, resolve_dagster_type(List[String]), ['obj_name']
            )


@nettest
def test_s3_intermediate_store_composite_types_with_custom_serializer_for_inner_type(s3_bucket):
    run_id = make_new_run_id()

    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket=s3_bucket)
    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_object(
                ['foo', 'bar'], context, resolve_dagster_type(List[LowercaseString]), ['list'],
            )
            assert intermediate_store.has_object(context, ['list'])
            assert intermediate_store.get_object(
                context, resolve_dagster_type(List[Bool]), ['list']
            ).obj == ['foo', 'bar']

        finally:
            intermediate_store.rm_object(context, ['foo'])


@nettest
def test_s3_intermediate_store_with_custom_serializer(s3_bucket):
    run_id = make_new_run_id()

    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket=s3_bucket)

    with yield_empty_pipeline_context(run_id=run_id) as context:
        try:
            intermediate_store.set_object('foo', context, LowercaseString, ['foo'])

            assert (
                intermediate_store.object_store.s3.get_object(
                    Bucket=intermediate_store.object_store.bucket,
                    Key='/'.join([intermediate_store.root] + ['foo']),
                )['Body']
                .read()
                .decode('utf-8')
                == 'FOO'
            )

            assert intermediate_store.has_object(context, ['foo'])
            assert intermediate_store.get_object(context, LowercaseString, ['foo']).obj == 'foo'
        finally:
            intermediate_store.rm_object(context, ['foo'])


@nettest
def test_s3_pipeline_with_custom_prefix(s3_bucket):
    s3_prefix = 'custom_prefix'

    pipe = define_inty_pipeline(should_throw=False)
    environment_dict = {
        'storage': {'s3': {'config': {'s3_bucket': s3_bucket, 's3_prefix': s3_prefix}}}
    }

    pipeline_run = PipelineRun(pipeline_name=pipe.name, environment_dict=environment_dict)
    instance = DagsterInstance.ephemeral()

    result = execute_pipeline(pipe, environment_dict=environment_dict,)
    assert result.success

    execution_plan = create_execution_plan(pipe, environment_dict)
    with scoped_pipeline_context(
        execution_plan, environment_dict, pipeline_run, instance,
    ) as context:
        store = S3IntermediateStore(
            run_id=result.run_id,
            s3_bucket=s3_bucket,
            s3_prefix=s3_prefix,
            s3_session=context.scoped_resources_builder.build(required_resource_keys={'s3'}).s3,
        )
        assert store.root == '/'.join(['custom_prefix', 'storage', result.run_id])
        assert store.get_intermediate(context, 'return_one.compute', Int).obj == 1
        assert store.get_intermediate(context, 'add_one.compute', Int).obj == 2


@nettest
def test_s3_intermediate_store_with_custom_prefix(s3_bucket):
    run_id = make_new_run_id()

    intermediate_store = S3IntermediateStore(
        run_id=run_id, s3_bucket=s3_bucket, s3_prefix='custom_prefix'
    )
    assert intermediate_store.root == '/'.join(['custom_prefix', 'storage', run_id])

    try:
        with yield_empty_pipeline_context(run_id=run_id) as context:

            intermediate_store.set_object(True, context, RuntimeBool, ['true'])

            assert intermediate_store.has_object(context, ['true'])
            assert intermediate_store.uri_for_paths(['true']).startswith(
                's3://%s/custom_prefix' % s3_bucket
            )

    finally:
        intermediate_store.rm_object(context, ['true'])


@nettest
def test_s3_intermediate_store(s3_bucket):
    run_id = make_new_run_id()
    run_id_2 = make_new_run_id()

    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket=s3_bucket)
    assert intermediate_store.root == '/'.join(['dagster', 'storage', run_id])

    intermediate_store_2 = S3IntermediateStore(run_id=run_id_2, s3_bucket=s3_bucket)
    assert intermediate_store_2.root == '/'.join(['dagster', 'storage', run_id_2])

    try:
        with yield_empty_pipeline_context(run_id=run_id) as context:

            intermediate_store.set_object(True, context, RuntimeBool, ['true'])

            assert intermediate_store.has_object(context, ['true'])
            assert intermediate_store.get_object(context, RuntimeBool, ['true']).obj is True
            assert intermediate_store.uri_for_paths(['true']).startswith('s3://')

            intermediate_store_2.copy_object_from_run(context, run_id, ['true'])
            assert intermediate_store_2.has_object(context, ['true'])
            assert intermediate_store_2.get_object(context, RuntimeBool, ['true']).obj is True
    finally:
        intermediate_store.rm_object(context, ['true'])
        intermediate_store_2.rm_object(context, ['true'])


class CsvSerializationStrategy(SerializationStrategy):
    def __init__(self):
        super(CsvSerializationStrategy, self).__init__(
            "csv_strategy", read_mode="r", write_mode="w"
        )

    def serialize(self, value, write_file_obj):
        fieldnames = value[0]
        writer = csv.DictWriter(write_file_obj, fieldnames)
        writer.writeheader()
        writer.writerows(value)

    def deserialize(self, read_file_obj):
        reader = csv.DictReader(read_file_obj)
        return LessSimpleDataFrame([row for row in reader])


@usable_as_dagster_type(
    name="LessSimpleDataFrame",
    description=("A naive representation of a data frame, e.g., as returned by " "csv.DictReader."),
    serialization_strategy=CsvSerializationStrategy(),
)
class LessSimpleDataFrame(list):
    pass


def test_custom_read_write_mode(s3_bucket):
    run_id = make_new_run_id()
    intermediate_store = S3IntermediateStore(run_id=run_id, s3_bucket=s3_bucket)
    data_frame = [OrderedDict({'foo': '1', 'bar': '1'}), OrderedDict({'foo': '2', 'bar': '2'})]
    try:
        with yield_empty_pipeline_context(run_id=run_id) as context:
            intermediate_store.set_object(
                data_frame, context, resolve_dagster_type(LessSimpleDataFrame), ['data_frame']
            )

            assert intermediate_store.has_object(context, ['data_frame'])
            assert (
                intermediate_store.get_object(
                    context, resolve_dagster_type(LessSimpleDataFrame), ['data_frame']
                ).obj
                == data_frame
            )
            assert intermediate_store.uri_for_paths(['data_frame']).startswith('s3://')

    finally:
        intermediate_store.rm_object(context, ['data_frame'])



from botocore.exceptions import ClientError

from dagster import Field, check, resource
from dagster.core.storage.file_cache import FileCache

from .file_manager import S3FileHandle
from .utils import create_s3_session


class S3FileCache(FileCache):
    def __init__(self, s3_bucket, s3_key, s3_session, overwrite=False):
        super(S3FileCache, self).__init__(overwrite=overwrite)

        self.s3_bucket = s3_bucket
        self.s3_key = s3_key
        self.s3 = s3_session

    def has_file_object(self, file_key):
        check.str_param(file_key, 'file_key')
        try:
            self.s3.get_object(Bucket=self.s3_bucket, Key=self.get_full_key(file_key))
        except ClientError:
            return False
        return True

    def get_full_key(self, file_key):
        return '{base_key}/{file_key}'.format(base_key=self.s3_key, file_key=file_key)

    def write_file_object(self, file_key, source_file_object):
        check.str_param(file_key, 'file_key')

        self.s3.put_object(
            Body=source_file_object, Bucket=self.s3_bucket, Key=self.get_full_key(file_key)
        )
        return self.get_file_handle(file_key)

    def get_file_handle(self, file_key):
        check.str_param(file_key, 'file_key')
        return S3FileHandle(self.s3_bucket, self.get_full_key(file_key))


@resource(
    {
        'bucket': Field(str),
        'key': Field(str),
        'overwrite': Field(bool, is_required=False, default_value=False),
    }
)
def s3_file_cache(init_context):
    return S3FileCache(
        s3_bucket=init_context.resource_config['bucket'],
        s3_key=init_context.resource_config['key'],
        overwrite=init_context.resource_config['overwrite'],
        # TODO: resource dependencies
        s3_session=create_s3_session(),
    )



import os
import re
import subprocess

import click
import pytest
from click.testing import CliRunner
from dagster_aws.cli.cli import (
    DAGSTER_HOME_ERROR,
    ensure_requirements,
    exit_gracefully,
    get_dagster_home,
    init,
    remove_ssh_key,
)
from moto import mock_ec2, mock_rds, mock_secretsmanager

from dagster import seven

TEST_PEM_PRIVATE_KEY = b'''-----BEGIN RSA PRIVATE KEY-----
MIIEowIBAAKCAQEAkMDeEQM0+z5OK15HPo/HFWA2deTOr9jFCT429CXJ1jEpcgyVixvVRXRqKrTi
otbRtqwzf5ZLi1/Jl04E5i4xtOvchgV4EHsBUy2pfv2AptBk0MWglviZrlqsNqtCB5JrNVo1dRU0
2cKEytaFZpkB5PYpVyts+XqEy4rADsG42FSAgTmSIS1YW3ts94OCuV2dTs3Rn2YVOE08Yl+Jzw/h
tifgrmSioIIZ82t/x02dgitZ+lhu7SSSuC1bNMXIiX5xd8Kt/eBZLV6cOO6s8jPqfokgADkCe/B2
0k0W2+croulAaGYyaFkzgfGclpYwh48j4tN1GAA+L/73jz4+m0MpBwIDAQABAoIBAQCMoTOdJ5Zz
eIaND5z7pMdvdvex9cbvhBN49V6CN6wtUbnIOX2XlkTOo1j5gJ6mQTmg1Q97JtLh5v985alQzxQv
hu3JrsqEiFwzKSuAWlyAs/kd8sIfqgn0H9crjZoqrQ9G2AsTYrTj0l0ciukzUfdLL/Pd4iyO+37E
GFqae45Pa9PrqLty/MwMpetRMxBd/ga76tbiIlskzJ0TWRN1n9vzmjxZsv7wvnLB6Lb9E2M7IbcF
4jT59PO9PDikyxXSCggtZ8zTOwwfJ0zT1ca0I8V5fAy8+8/gEipPS2w4FDXHDNZYTFjPkAbIcVjk
VOs1gIXnVU9Udx5ZbqYyAc6dIXERAoGBAP+3Mg7AK54UbxDHE+vZ90kWf4zaNCrzlIa6hGO4UGfi
Hu4Q2yi7N4/PFPzqK2fiDmF3KcADDFoalpBjgrkuCwN/mzI5DTmXezC+ybwWDHv5LOpxKSt6jI3k
Cl7gPluSFlpO1PZLqvCoD/2sYEC5UMOtdm3mAbdRmAe0DZWy8KijAoGBAJDqFHrIotedciYrkbwy
ZvH8U4AatgHRNwRyM+pOSPh1mGIsxCLU7/JNNO85Lsj7sY9vrUbNCoUMh3tW7aIPqSM7984BN3OH
j5BDI2UNKjMU2I+/a8hSevBpVAiuXwZdWho2R9X+bOQEJ5/XSCcKb2IEsxnsvUZuJ8PDIUgydNBN
AoGAChAYfTIcxUzCjiN3ajmanJqxDEbt9J6/QooGebIgH+ZrFy08op6zcgpRJh7E9IKV1EoPL4ov
K0COmkIAO+9O6aVU4yYRmimW5HUtlSQ8+4fzt4oad6aL722A5vjmG05laMpfYQ3bPTsv3eixpPpl
7j1eQrbhbGheljcErdZr2AcCgYBkeDX8rq/f50Pl3N0sapFfAj7uGRbJCcEoLvl5YA9NW1Jr9neB
Yrp89jRWCwI5y+7d8GJlLPE3L2mbBLi6XDs6t2/l/ofwbDMHpNScUdVJFxHSD7ftaencVloxrpsp
MX2voKSLY9sg9xR1yG84yg5RGcsGlVDgzKx7GAUyJfFBHQKBgB+xoOvKJRalORPxryV5g37q/0C9
/HqpGCbVxGTuKdq9aASuOae3Lxf6b7NyLVjPLnhdOCwddi8tFbPoZa+3/EBSX1VPVxXAJ/FpkIbW
/HUvDDKSnDcBzS1/ufsTgRO4OqtTrj9idqoyAPar4BdC1ABDbL4OOKP9QAPrvjefBZtt
-----END RSA PRIVATE KEY-----'''


def test_ensure_requirements():
    runner = CliRunner()
    with runner.isolated_filesystem() as tmp_dir:

        @click.command()
        def test():
            ensure_requirements(tmp_dir)

        result = runner.invoke(test)
        assert 'No requirements.txt found, creating' in result.output

        with open(os.path.join(tmp_dir, 'requirements.txt'), 'w') as f:
            f.write('dagster\ndagit\n')

        result = runner.invoke(test)
        assert 'Found existing requirements.txt' in result.output

        with open(os.path.join(tmp_dir, 'requirements.txt'), 'w') as f:
            f.write('dagster\n')

        result = runner.invoke(test, input='n\n')
        assert 'Could not find dagit' in result.output
        assert result.exit_code == 1

        result = runner.invoke(test, input='y\n')
        assert 'Could not find dagit' in result.output
        assert result.exit_code == 0

        with open(os.path.join(tmp_dir, 'requirements.txt'), 'w') as f:
            f.write('dagit\n')

        result = runner.invoke(test, input='n\n')
        assert 'Could not find dagster' in result.output
        assert result.exit_code == 1

        result = runner.invoke(test, input='y\n')
        assert 'Could not find dagster' in result.output
        assert result.exit_code == 0


def test_remove_ssh_key():
    # Ensure SSH is running
    subprocess.call(['ssh-agent', '-s'])

    test_key_path = os.path.join(seven.get_system_temp_directory(), 'test.pem')
    with open(test_key_path, 'wb') as f:
        f.write(TEST_PEM_PRIVATE_KEY)

    os.chmod(test_key_path, 0o600)

    subprocess.call(['ssh-add', '-D'])
    assert remove_ssh_key('does/not/matter')

    subprocess.call(['ssh-add', test_key_path])
    assert not remove_ssh_key('/key/does/not/exist.pem')
    assert remove_ssh_key(test_key_path)


def test_get_dagster_home(capsys, tmp_path):
    old_env = os.getenv('DAGSTER_HOME')
    if old_env is not None:
        del os.environ['DAGSTER_HOME']

    with pytest.raises(SystemExit) as exc_info:
        get_dagster_home()

    captured = capsys.readouterr()

    assert exc_info.value.code == 1
    assert DAGSTER_HOME_ERROR in captured.err

    path = str(tmp_path)
    os.environ['DAGSTER_HOME'] = path
    assert get_dagster_home() == path

    captured = capsys.readouterr()
    assert ('Found DAGSTER_HOME in environment at: {path}'.format(path=path)) in captured.out

    if old_env is not None:
        os.environ['DAGSTER_HOME'] = old_env


def test_exit_gracefully(capsys):
    with pytest.raises(SystemExit) as exc_info:
        exit_gracefully(None, None)

    captured = capsys.readouterr()

    assert exc_info.value.code == 1
    assert 'Command killed by keyboard interrupt, quitting' in captured.out


@mock_ec2
@mock_rds
@mock_secretsmanager
def test_init(tmp_path):
    region = 'us-east-1'
    ami_id = 'ami-03cf127a'
    keypair_name = 'foobar-keypair'
    instance_type = 't3.large'
    instance_name = 'foobar'
    use_rds = 'N'

    stdin = '\n'.join(
        [
            region,
            '',  # VPC ID
            '',  # security group
            ami_id,
            keypair_name,
            instance_type,
            instance_name,
            use_rds,
        ]
    )

    runner = CliRunner()
    result = runner.invoke(init, [], env={'DAGSTER_HOME': str(tmp_path)}, input=stdin)
    assert result.exit_code == 0

    assert (
        'Found DAGSTER_HOME in environment at: {tmp_path}'.format(tmp_path=str(tmp_path))
        in result.output
    )
    assert 'Security group created' in result.output
    assert (
        'Key pair {keypair_name} created and saved to local file {tmp_path}/keys/foobar-keypair.pem!'.format(
            keypair_name=keypair_name, tmp_path=str(tmp_path)
        )
        in result.output
    )
    assert re.search(r'dagit EC2 instance i\-.*? launched!', result.output)

    assert (
        'Saved EC2 Configuration configuration to {tmp_path}/dagster-aws-config.yaml'.format(
            tmp_path=str(tmp_path)
        )
        in result.output
    )
    assert region in result.output
    assert ami_id in result.output
